"""
This is a helper class to provide some common functionality around HTTP
client operations.
"""
import asyncio
from collections import namedtuple
import logging
import ssl

import aiohttp
import httpx
import sys
import vcr
import yaml

# This library used by aiohttp 3.8 needs its logging turned down
logging.getLogger("charset_normalizer").setLevel(logging.ERROR)
logging.getLogger("httpx").setLevel(logging.WARNING)  # STM-10584

# STM-3012 found that enabling enable_cleanup_closed reduced memory leaks.
# STM-10143 found that disabling this fixed a memory leak in python >= 3.11.2.
_enable_cleanup_closed = False
if sys.version_info < (3, 11, 2):
    _enable_cleanup_closed = True


class HttpClient(object):
    """Helper class to simplify HTTP operations on an open session."""

    def __init__(self,
                 url,
                 verify_ssl=True,
                 username=None,
                 password=None,
                 auth=None,
                 cert=None,
                 key=None,
                 user_agent=None,
                 connection_limit=None,
                 conn_timeout=None,
                 read_timeout=None,
                 record=None):
        """Initialize the session pool and connector objects."""
        work_around_aiohttp_3535()
        timeout = aiohttp.ClientTimeout(connect=conn_timeout or 30,
                                        sock_read=read_timeout or 300)
        if verify_ssl:
            conn = aiohttp.TCPConnector(limit_per_host=connection_limit,
                                        enable_cleanup_closed=_enable_cleanup_closed)
            self._session = aiohttp.ClientSession(connector=conn, timeout=timeout)
        else:
            ssl_ctx = ssl._create_unverified_context()
            try:
                try:
                    ssl_ctx.minimum_version = ssl.TLSVersion.TLSv1  # STM-6823
                except AttributeError:
                    pass  # Python 3.6
                ssl_ctx.set_ciphers('DEFAULT@SECLEVEL=1')  # STM-4531
            except ssl.SSLError:
                pass  # some platforms don't allow this, ok
            if cert and key:
                ssl_ctx.load_cert_chain(cert, key)
                conn = aiohttp.TCPConnector(
                    ssl_context=ssl_ctx,
                    limit_per_host=connection_limit,
                    enable_cleanup_closed=_enable_cleanup_closed)
            else:
                conn = aiohttp.TCPConnector(
                    ssl_context=ssl_ctx,
                    limit_per_host=connection_limit,
                    enable_cleanup_closed=_enable_cleanup_closed)
            self._session = aiohttp.ClientSession(
                connector=conn,
                timeout=timeout,
                cookie_jar=aiohttp.CookieJar(unsafe=True))
        self._url = url
        self.auth = auth
        if username:
            if password:
                self.auth = aiohttp.BasicAuth(username, password)
            else:
                self.auth = aiohttp.BasicAuth(username)
        self.user_agent = user_agent or 'preseem-netmeta-agent'

        self.vcr_cassette = None
        if record:  # record all http transactions to the provided name
            self.record(record)
        self.http2_client = None
        self.http2_record = None

    def record(self, name):
        if not self.vcr_cassette:
            vcr_log = logging.getLogger("vcr")
            vcr_log.setLevel(logging.WARNING)
            self.vcr = vcr.VCR(record_mode='all', cassette_library_dir='/tmp')
            self.vcr_cm = self.vcr.use_cassette(name)
            self.vcr_cassette = self.vcr_cm.__enter__()

    async def http2_request(self, path, num_retries=5, **kwargs):
        """Make a http2 request.  Returns the httpx response object."""
        if not self.http2_client:
            self.http2_client = httpx.AsyncClient(http2=True,
                                                  verify=False,
                                                  timeout=httpx.Timeout(5.0, read=30.0))
        url = '{}{}'.format(self._url, path)
        err = None
        for retry_num in range(num_retries + 1):
            if retry_num > 0:
                await asyncio.sleep(2**retry_num)
                if retry_num > 1:
                    # don't log on the first retry, let it fail once silently
                    logging.warning('Retry request to {}, last error {}'.format(
                        url, err))
            method = 'POST' if kwargs.get('data') or kwargs.get('json') or kwargs.get(
                'content') else 'GET'
            try:
                # We want to just return the request to the caller, for now.
                response = await self.http2_client.request(method, url, **kwargs)
                if self.vcr_cassette is not None:
                    if self.http2_record is None:
                        self.http2_record = []
                    self.http2_record.append({
                        'request': {
                            'url': url,
                            **kwargs
                        },
                        'response': {
                            'status_code': response.status_code,
                            'reason_phrase': response.reason_phrase,
                            'http_version': response.http_version,
                            'url': str(response.url),
                            'headers': dict(response.headers),
                            'content': response.content,
                        }
                    })
                return response
            except asyncio.CancelledError:
                raise
            except asyncio.TimeoutError:
                err = 'TimeoutError'
            except httpx.RequestError as e:
                err = e
        raise RuntimeError(f'Error fetching {url}: {err}')

    async def http2_stream(self, path, **kwargs):
        """Make a http2 stream request.  Returns the httpx response object."""
        url = '{}{}'.format(self._url, path)
        method = 'POST' if kwargs.get('data') or kwargs.get('json') or kwargs.get(
            'content') else 'GET'
        if not self.http2_client:
            self.http2_client = httpx.AsyncClient(http2=True,
                                                  verify=False,
                                                  timeout=httpx.Timeout(5.0, read=30.0))
        req = self.http2_client.build_request(method, url, **kwargs)
        r = await self.http2_client.send(req, stream=True)
        r.preseem_client = self
        if self.vcr_cassette is not None:  # record this session
            if self.http2_record is None:
                self.http2_record = []
            self.http2_record.append({
                'request': {
                    'url': url,
                    **kwargs
                },
                'response': {
                    'status_code': r.status_code,
                    'reason_phrase': r.reason_phrase,
                    'http_version': r.http_version,
                    'url': str(r.url),
                    'headers': dict(r.headers),
                },
                'stream': []
            })
        return r

    def record_http2_stream_data(self, r, data):
        """Helper to facilitate recording data for http streams.  This is
           needed because I haven't found a good way to do it transparently."""
        if self.vcr_cassette is not None:
            cur = self.http2_record[-1]
            if 'stream' in cur and cur['request']['url'] == r.url:
                cur['stream'].append(data)

    async def close(self):
        if self.http2_client:
            await self.http2_client.aclose()
        if self.vcr_cassette is not None:
            self.vcr_cm.__exit__()
            self.vcr_cassette = None
        if self._session:
            await self._session.close()

    async def request(self,
                      path,
                      num_retries=5,
                      handler=None,
                      response_encoding=None,
                      **kwargs):
        """Post a request to the session."""
        hdrs = {'user-agent': self.user_agent}
        hdrs.update(kwargs.get('headers') or {})
        if 'headers' in kwargs:
            del kwargs['headers']
        if self.auth and 'auth' not in kwargs:
            kwargs['auth'] = self.auth
        url = '{}{}'.format(self._url, path)
        err = None
        for retry_num in range(num_retries + 1):
            if retry_num > 0:
                await asyncio.sleep(2**retry_num)
                if retry_num > 1:
                    # don't log on the first retry, let it fail once silently
                    logging.warning('Retry request to {}, last error {}'.format(
                        url, err))
            try:
                method = 'POST' if kwargs.get('data') or kwargs.get('json') else 'GET'
                if 'method' in kwargs:
                    method = kwargs['method']
                    del kwargs['method']
                async with self._session.request(method, url, headers=hdrs,
                                                 **kwargs) as r:
                    if handler:
                        # Allow a custom response handler to be passed in
                        return await handler(r)
                    if 200 <= r.status < 300:
                        if r.content_type == 'application/json':
                            if response_encoding:
                                return await r.json(encoding=response_encoding)
                            else:
                                return await r.json()
                        else:
                            if response_encoding:
                                return await r.text(encoding=response_encoding)
                            else:
                                return await r.text()
                    msg = await r.text()
                    err = 'Status code={} text={}'.format(r.status, msg)
                    if r.status >= 500:
                        err = r.text
                        continue
                    r.raise_for_status()
            except asyncio.TimeoutError:
                err = 'TimeoutError'
            except aiohttp.ServerDisconnectedError:
                # This is just a regular occurrence.
                # We only log an error on the second retry to avoid this.
                err = 'ServerDisconnectedError'
            except aiohttp.ClientConnectionError:
                err = 'ClientConnectionError'
            except aiohttp.ServerDisconnectedError:
                err = 'ServerDisconnectedError'
        raise RuntimeError('Error fetching {}: {}'.format(url, err))


def ignore_aiohttp_ssl_eror(loop, aiohttpversion='3.8.1'):
    """Ignore aiohttp #3535 issue with SSL data after close
    There appears to be an issue on Python 3.7 and aiohttp SSL that throws a
    ssl.SSLError fatal error (ssl.SSLError: [SSL: KRB5_S_INIT] application data
    after close notify (_ssl.c:2609)) after we are already done with the
    connection. See GitHub issue aio-libs/aiohttp#3535

    Given a loop, this sets up a exception handler that ignores this specific
    exception, but passes everything else on to the previous exception handler
    this one replaces.

    If the current aiohttp version is not exactly equal to aiohttpversion
    nothing is done, assuming that the next version will have this bug fixed.
    This can be disabled by setting this parameter to None
    """
    if aiohttpversion is not None and aiohttp.__version__ != aiohttpversion:
        return

    orig_handler = loop.get_exception_handler() or loop.default_exception_handler

    def ignore_ssl_error(loop, context):
        if context.get('message') == 'SSL error in data received':
            # validate we have the right exception, transport and protocol
            exception = context.get('exception')
            protocol = context.get('protocol')
            if (isinstance(exception, ssl.SSLError)
                    and exception.reason == 'KRB5_S_INIT'
                    and isinstance(protocol, asyncio.sslproto.SSLProtocol)
                    and isinstance(protocol._app_protocol,
                                   aiohttp.client_proto.ResponseHandler)):
                if loop.get_debug():
                    asyncio.log.logger.debug('Ignoring aiohttp SSL KRB5_S_INIT error')
                return
        try:
            orig_handler(loop, context)
        except TypeError:
            orig_handler(context)

    loop.set_exception_handler(ignore_ssl_error)


_worked_around_aiohttp_3535 = False


def work_around_aiohttp_3535():
    global _worked_around_aiohttp_3535
    if _worked_around_aiohttp_3535:
        return
    try:
        ignore_aiohttp_ssl_eror(asyncio.get_running_loop())
    except AttributeError:  # predates python 3.7
        ignore_aiohttp_ssl_eror(asyncio.get_event_loop())
    finally:
        _worked_around_aiohttp_3535 = True


def dict2fs(d):
    return frozenset(
        (k, frozenset(v.items()) if isinstance(v, dict) else v) for k, v in d.items())


class ReplayHttpClient(HttpClient):
    """Special HTTP client that replays a capture file.  Currently http2 only,
       we use vcrpy to replay other http."""
    file = None
    http2_client = None
    _session = None

    def __init__(self, url, file=None):
        super().__init__(url)
        self._url = url
        file = file or self.file
        if file:
            with open(file) as f:  # we assume caller set the file class var
                data = yaml.safe_load(f.read())
            self.data = {}
            for rec in data:
                self.data[dict2fs(rec['request'])] = rec
        self.vcr_cassette = None

    async def http2_request(self, path, num_retries=None, **kwargs):
        """Make a http2 request.  Returns a httpx-compatible response object."""
        url = f'{self._url}{path}'
        key = dict2fs({'url': url, **kwargs})
        rec = self.data.get(key)
        if not rec:
            raise RuntimeError(f'No matching record for {key}')
        # return an object that will behave like a httpx.Response
        nt = namedtuple('Response', rec['response'].keys())
        return nt(*rec['response'].values())

    class FakeStreamResponse:
        """async iterator that works like the httpx stream response."""

        def __init__(self, chunks):
            self.chunks = chunks

        async def aclose(self):
            pass

        async def aiter_bytes(self):
            for chunk in self.chunks:
                yield chunk

    async def http2_stream(self, path, **kwargs):
        """Make a http2 stream request.  Returns a fake response object."""
        url = '{}{}'.format(self._url, path)
        key = dict2fs({'url': url, **kwargs})
        rec = self.data.get(key)
        if not rec or 'stream' not in rec:
            raise RuntimeError(f'No matching stream record for {key}')
        r = self.FakeStreamResponse(rec['stream'])
        for k, v in rec['response'].items():
            setattr(r, k, v)
        return r

    def record_http2_stream_data(self, r, data):
        pass



"""
Proxy classes for network-poller protobuf messages.
These classes allow bounds checking and error handling to be centralized.
"""
import base64
from collections import namedtuple
from datetime import timezone
import json
import logging
from preseem_protobuf.network_poller import network_poller_pb2
from uuid import UUID

from google.protobuf.json_format import MessageToDict


class NetworkElementUpdateException(Exception):
    """Allows a poll operation to raise an error for the NetworkElement."""

    def __init__(self, error, msg):
        self.error = error
        super().__init__(msg)


class ErrorContext:
    """Container for holding errors during protobuf operations."""
    Error = namedtuple('Error', ('code', 'msg', 'args'))

    def __init__(self):
        self.errors = []

    def add(self, code, msg, **kwargs):
        """Add an error."""
        self.errors.append(self.Error(code, msg, kwargs))

    def diff(self, other):
        """Return a list of errors that are in other but not in this context."""
        return [x for x in other.errors if x not in self.errors]


class Repeated:

    def __init__(self, py_type, pb_field, errctx, checker=None):
        self.py_type = py_type
        self.pb_field = pb_field
        self.errctx = errctx
        self.checker = checker

    def __iter__(self):
        return iter([self.py_type(pb) if self.py_type else pb for pb in self.pb_field])

    def __len__(self):
        return len(self.pb_field)

    def __getitem__(self, i):
        return self.py_type(self.pb_field[i]) if self.py_type else self.pb_field[i]

    def __delitem__(self, i):
        del self.pb_field[i]

    def add(self):
        assert self.py_type  # this method only works for message fields
        pb = self.pb_field.add()
        return self.py_type(pb, errctx=self.errctx)

    def append(self, value):
        if not self.checker or self.checker(value):
            if self.py_type and isinstance(value, self.py_type):
                self.pb_field.append(value.pb)
            else:
                self.pb_field.append(value)


class BaseWrapperClass:
    """A base class for protobuf wrapper classes"""

    def __init__(self, pb=None, errctx=None):
        self.pb = pb if pb else self.PBTYPE()
        self.errctx = errctx

    def __eq__(self, other):
        """define equivalence as the equivalence of the wrapped protobuf."""
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.pb == other.pb

    def __repr__(self):
        """repr function for pb wrapper classes.  Requires _attrs to be defined."""
        attr_values = [(attr, getattr(self, attr)) for attr in self._attrs]
        attributes_str = ', '.join(f"{attr}={value!r}" for attr, value in attr_values)
        return f"{type(self).__name__}({attributes_str})"

    def pretty(self):
        """return a pretty-printed string representing this full message."""

        def unencode(d):
            for a, v in d.items():
                if a.endswith('uuid'):
                    try:
                        d[a] = str(UUID(bytes=base64.b64decode(v)))
                    except ValueError:
                        d[a] = v
                elif a == 'poller_hash':
                    v = base64.b64decode(v)
                    if len(v) == 32:
                        # assume sha256 poller hash
                        d[a] = v.hex()
                    else:
                        # likely an intermediate value, show as a string
                        # (can't convert bytes to json)
                        d[a] = v.decode()
                elif isinstance(v, dict):
                    unencode(v)
                elif isinstance(v, list):
                    for x in v:
                        if isinstance(x, dict):
                            unencode(x)

        m = MessageToDict(self.pb, preserving_proto_field_name=True)
        unencode(m)
        return json.dumps(m, indent=2)


def pb_wrapper(pbtype):
    """Class decorator that sets the PBTYPE and creates an _attrs member."""

    def class_decorator(cls):
        setattr(cls, 'PBTYPE', pbtype)
        cls._attrs = [
            name for name, value in vars(cls).items() if isinstance(value, property)
        ]
        return cls

    return class_decorator


@pb_wrapper(network_poller_pb2.NetworkElement.Link)
class Link(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        self.bridged_mac_addresses = Repeated(None, self.pb.bridged_mac_addresses,
                                              self.errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def oper_up(self):
        return self.pb.oper_up if self.pb.HasField('oper_up') else None

    @oper_up.setter
    def oper_up(self, value):
        if value is not None:
            self.pb.oper_up = value
        else:
            self.pb.ClearField('oper_up')

    @property
    def mac_address(self):
        return self.pb.mac_address

    @mac_address.setter
    def mac_address(self, value):
        if value is not None:
            self.pb.mac_address = value

    @property
    def in_octets(self):
        return self.pb.in_octets if self.pb.HasField('in_octets') else None

    @in_octets.setter
    def in_octets(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_octets = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.in_octets',
                                value=value,
                                metric_name='NetworkElement.Link.in_octets')
        else:
            self.pb.ClearField('in_octets')

    @property
    def in_ucast_pkts(self):
        return self.pb.in_ucast_pkts if self.pb.HasField('in_ucast_pkts') else None

    @in_ucast_pkts.setter
    def in_ucast_pkts(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_ucast_pkts = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.in_ucast_pkts',
                                value=value,
                                metric_name='NetworkElement.Link.in_ucast_pkts')
        else:
            self.pb.ClearField('in_ucast_pkts')

    @property
    def in_nucast_pkts(self):
        return self.pb.in_nucast_pkts if self.pb.HasField('in_nucast_pkts') else None

    @in_nucast_pkts.setter
    def in_nucast_pkts(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_nucast_pkts = value
            else:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.in_nucast_pkts',
                                value=value,
                                metric_name='NetworkElement.Link.in_nucast_pkts')
        else:
            self.pb.ClearField('in_nucast_pkts')

    @property
    def in_discards(self):
        return self.pb.in_discards if self.pb.HasField('in_discards') else None

    @in_discards.setter
    def in_discards(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_discards = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.in_discards',
                                value=value,
                                metric_name='NetworkElement.Link.in_discards')
        else:
            self.pb.ClearField('in_discards')

    @property
    def in_errors(self):
        return self.pb.in_errors if self.pb.HasField('in_errors') else None

    @in_errors.setter
    def in_errors(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_errors = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.in_errors',
                                value=value,
                                metric_name='NetworkElement.Link.in_errors')
        else:
            self.pb.ClearField('in_errors')

    @property
    def in_unknown_protos(self):
        return self.pb.in_unknown_protos if self.pb.HasField(
            'in_unknown_protos') else None

    @in_unknown_protos.setter
    def in_unknown_protos(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_unknown_protos = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.Link.in_unknown_protos',
                    value=value,
                    metric_name=' NetworkElement.Link.in_unknown_protos')
        else:
            self.pb.ClearField('in_unknown_protos')

    @property
    def out_octets(self):
        return self.pb.out_octets if self.pb.HasField('out_octets') else None

    @out_octets.setter
    def out_octets(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_octets = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.out_octets',
                                value=value,
                                metric_name="NetworkElement.Link.out_octets")
        else:
            self.pb.ClearField('out_octets')

    @property
    def out_ucast_pkts(self):
        return self.pb.out_ucast_pkts if self.pb.HasField('out_ucast_pkts') else None

    @out_ucast_pkts.setter
    def out_ucast_pkts(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_ucast_pkts = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.out_ucast_pkts',
                                value=value,
                                metric_name='NetworkElement.Link.out_ucast_pkts')
        else:
            self.pb.ClearField('out_ucast_pkts')

    @property
    def out_nucast_pkts(self):
        return self.pb.out_nucast_pkts if self.pb.HasField('out_nucast_pkts') else None

    @out_nucast_pkts.setter
    def out_nucast_pkts(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_nucast_pkts = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.out_nucast_pkts',
                                value=value,
                                metric_name='NetworkElement.Link.out_nucast_pkts')
        else:
            self.pb.ClearField('out_nucast_pkts')

    @property
    def out_discards(self):
        return self.pb.out_discards if self.pb.HasField('out_discards') else None

    @out_discards.setter
    def out_discards(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_discards = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.out_discards',
                                value=value,
                                metric_name='NetworkElement.Link.out_discards')
        else:
            self.pb.ClearField('out_discards')

    @property
    def out_errors(self):
        return self.pb.out_errors if self.pb.HasField('out_errors') else None

    @out_errors.setter
    def out_errors(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_errors = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.out_errors',
                                value=value,
                                metric_name='NetworkElement.Link.out_errors')
        else:
            self.pb.ClearField('out_errors')

    @property
    def out_retransmits(self):
        return self.pb.out_retransmits if self.pb.HasField('out_retransmits') else None

    @out_retransmits.setter
    def out_retransmits(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_retransmits = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Link.out_retransmits',
                                value=value,
                                metric_name='NetworkElement.Link.out_retransmits')
            else:
                self.pb.ClearField('out_retransmits')

    @property
    def poller_hash(self):
        return self.pb.poller_hash if self.pb.HasField('poller_hash') else None

    @poller_hash.setter
    def poller_hash(self, value):
        if value is not None:
            self.pb.poller_hash = value

    def add_bridged_mac(self, mac):
        if mac.startswith('33:33:') or mac == '00:00:00:00:00:00':
            return  # ignore IPv6 multicast MACs (PPA-1321) and zero MACs
        self.bridged_mac_addresses.append(mac)


@pb_wrapper(network_poller_pb2.NetworkElement.WirelessLink)
class WirelessLink(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def mac_address(self):
        return self.pb.mac_address

    @mac_address.setter
    def mac_address(self, value):
        if value is not None:
            self.pb.mac_address = value

    @property
    def distance(self):
        return self.pb.distance if self.pb.HasField('distance') else None

    @distance.setter
    def distance(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.distance = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.WirelessLink.distance',
                    value=value,
                    metric_name='NetworkElement.WirelessLink.distance')
        else:
            self.pb.ClearField('distance')

    @property
    def connected_time(self):
        return self.pb.connected_time if self.pb.HasField('connected_time') else None

    @connected_time.setter
    def connected_time(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.connected_time = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.WirelessLink.connected_time',
                    value=value,
                    metric_name='NetworkElement.WirelessLink.connected_time')
        else:
            self.pb.ClearField('connected_time')

    @property
    def noise_floor(self):
        return self.pb.noise_floor if self.pb.HasField('noise_floor') else None

    @noise_floor.setter
    def noise_floor(self, value):
        if value is not None:
            self.pb.noise_floor = value
        else:
            self.pb.ClearField('noise_floor')

    @property
    def rssi(self):
        return self.pb.rssi if self.pb.HasField('rssi') else None

    @rssi.setter
    def rssi(self, value):
        if value is not None:
            self.pb.rssi = value
        else:
            self.pb.ClearField('rssi')

    @property
    def snr(self):
        return self.pb.snr if self.pb.HasField('snr') else None

    @snr.setter
    def snr(self, value):
        if value is not None:
            self.pb.snr = value
        else:
            self.pb.ClearField('snr')

    @property
    def rx_link_rate(self):
        return self.pb.rx_link_rate if self.pb.HasField('rx_link_rate') else None

    @rx_link_rate.setter
    def rx_link_rate(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.rx_link_rate = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.WirelessLink.rx_link_rate',
                    value=value,
                    metric_name='NetworkElement.WirelessLink.rx_link_rate')
        else:
            self.pb.ClearField('rx_link_rate')

    @property
    def tx_link_rate(self):
        return self.pb.tx_link_rate if self.pb.HasField('tx_link_rate') else None

    @tx_link_rate.setter
    def tx_link_rate(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.tx_link_rate = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.WirelessLink.tx_link_rate',
                    value=value,
                    metric_name='NetworkElement.WirelessLink.tx_link_rate')
        else:
            self.pb.ClearField('tx_link_rate')

    @property
    def rx_mcs(self):
        return self.pb.rx_mcs if self.pb.HasField('rx_mcs') else None

    @rx_mcs.setter
    def rx_mcs(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.rx_mcs = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.WirelessLink.rx_mcs',
                                value=value,
                                metric_name=' NetworkElement.WirelessLink.rx_mcs')
        else:
            self.pb.ClearField('rx_mcs')

    @property
    def tx_mcs(self):
        return self.pb.tx_mcs if self.pb.HasField('tx_mcs') else None

    @tx_mcs.setter
    def tx_mcs(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.tx_mcs = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.WirelessLink.tx_mcs',
                                value=value,
                                metric_name='NetworkElement.WirelessLink.tx_mcs')
        else:
            self.pb.ClearField('tx_mcs')

    @property
    def rx_modulation_rate(self):
        return self.pb.rx_modulation_rate if self.pb.HasField(
            'rx_modulation_rate') else None

    @rx_modulation_rate.setter
    def rx_modulation_rate(self, value):
        if value is not None:
            self.pb.rx_modulation_rate = value
        else:
            self.pb.ClearField('rx_modulation_rate')

    @property
    def tx_modulation_rate(self):
        return self.pb.tx_modulation_rate if self.pb.HasField(
            'tx_modulation_rate') else None

    @tx_modulation_rate.setter
    def tx_modulation_rate(self, value):
        if value is not None:
            self.pb.tx_modulation_rate = value
        else:
            self.pb.ClearField('tx_modulation_rate')

    @property
    def rx_airtime(self):
        return self.pb.rx_airtime if self.pb.HasField('rx_airtime') else None

    @rx_airtime.setter
    def rx_airtime(self, value):
        if value is not None:
            if value >= 0:
                self.pb.rx_airtime = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for WirelessLink.rx_airtime',
                                value=value,
                                metric_name='WirelessLink.rx_airtime')
        else:
            self.pb.ClearField('rx_airtime')

    @property
    def tx_airtime(self):
        return self.pb.tx_airtime if self.pb.HasField('tx_airtime') else None

    @tx_airtime.setter
    def tx_airtime(self, value):
        if value is not None:
            if value >= 0:
                self.pb.tx_airtime = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for WirelessLink.tx_airtime',
                                value=value,
                                metric_name='WirelessLink.tx_airtime')
        else:
            self.pb.ClearField('tx_airtime')

    @property
    def amc(self):
        return self.pb.amc if self.pb.HasField('amc') else None

    @amc.setter
    def amc(self, value):
        if value is not None:
            if value >= 0 and value <= 100:
                self.pb.amc = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for WirelessLink.amc',
                                value=value,
                                metric_name='WirelessLink.amc')
        else:
            self.pb.ClearField('amc')

    @property
    def amq(self):
        return self.pb.amq if self.pb.HasField('amq') else None

    @amq.setter
    def amq(self, value):
        if value is not None:
            if value >= 0 and value <= 100:
                self.pb.amq = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for WirelessLink.amq',
                                value=value,
                                metric_name='WirelessLink.amq')
        else:
            self.pb.ClearField('amq')

    @property
    def ccq(self):
        return self.pb.ccq if self.pb.HasField('ccq') else None

    @ccq.setter
    def ccq(self, value):
        if value is not None:
            if value >= 0 and value <= 100:
                self.pb.ccq = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for WirelessLink.ccq',
                                value=value,
                                metric_name='WirelessLink.ccq')
        else:
            self.pb.ClearField('ccq')

    @property
    def poller_hash(self):
        return self.pb.poller_hash if self.pb.HasField('poller_hash') else None

    @poller_hash.setter
    def poller_hash(self, value):
        if value is not None:
            self.pb.poller_hash = value


@pb_wrapper(network_poller_pb2.NetworkElement.RadioStream)
class RadioStream(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        self.links = Repeated(WirelessLink, self.pb.links, self.errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def id(self):
        return self.pb.id

    @id.setter
    def id(self, value):
        if value is not None:
            self.pb.id = value

    @property
    def polarization(self):
        return self.pb.polarization if self.pb.HasField('polarization') else None

    @polarization.setter
    def polarization(self, value):
        if value is not None:
            self.pb.polarization = value
        else:
            self.pb.ClearField('polarization')

    @property
    def noise_floor(self):
        return self.pb.noise_floor if self.pb.HasField('noise_floor') else None

    @noise_floor.setter
    def noise_floor(self, value):
        if value is not None:
            self.pb.noise_floor = value
        else:
            self.pb.ClearField('noise_floor')


@pb_wrapper(network_poller_pb2.NetworkElement.RfAggregationGroup)
class RfAggregationGroup(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def name(self):
        return self.pb.name if self.pb.HasField('name') else None

    @name.setter
    def name(self, value):
        if value:
            self.pb.name = value
        else:
            self.pb.ClearField('name')

    @property
    def airtime(self):
        return self.pb.airtime if self.pb.HasField('airtime') else None

    @airtime.setter
    def airtime(self, value):
        if value is not None:
            if value >= 0:
                self.pb.airtime = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.RfAggregationGroup.airtime',
                    value=value,
                    metric_name='NetworkElement.RfAggregationGroup.airtime')
        else:
            self.pb.ClearField('airtime')

    @property
    def tx_packet_error_rate(self):
        return self.pb.tx_packet_error_rate if self.pb.HasField(
            'tx_packet_error_rate') else None

    @tx_packet_error_rate.setter
    def tx_packet_error_rate(self, value):
        if value is not None:
            if value >= 0:
                self.pb.tx_packet_error_rate = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.RfAggregationGroup.tx_packet_error_rate',
                    value=value,
                    metric_name='NetworkElement.RfAggregationGroup.tx_packet_error_rate'
                )
        else:
            self.pb.ClearField('tx_packet_error_rate')

    @property
    def rx_packet_error_rate(self):
        return self.pb.rx_packet_error_rate if self.pb.HasField(
            'rx_packet_error_rate') else None

    @rx_packet_error_rate.setter
    def rx_packet_error_rate(self, value):
        if value is not None:
            if value >= 0:
                self.pb.rx_packet_error_rate = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.RfAggregationGroup.rx_packet_error_rate',
                    value=value,
                    metric_name='NetworkElement.RfAggregationGroup.rx_packet_error_rate'
                )
        else:
            self.pb.ClearField('rx_packet_error_rate')

    @property
    def tx_interference_signal_ratio(self):
        return self.pb.tx_interference_signal_ratio if self.pb.HasField(
            'tx_interference_signal_ratio') else None

    @tx_interference_signal_ratio.setter
    def tx_interference_signal_ratio(self, value):
        if value is not None:
            self.pb.tx_interference_signal_ratio = value
        else:
            self.pb.ClearField('tx_interference_signal_ratio')

    @property
    def rx_interference_signal_ratio(self):
        return self.pb.rx_interference_signal_ratio if self.pb.HasField(
            'rx_interference_signal_ratio') else None

    @rx_interference_signal_ratio.setter
    def rx_interference_signal_ratio(self, value):
        if value is not None:
            self.pb.rx_interference_signal_ratio = value
        else:
            self.pb.ClearField('rx_interference_signal_ratio')


@pb_wrapper(network_poller_pb2.NetworkElement.Radio)
class Radio(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        self.streams = Repeated(RadioStream, self.pb.streams, self.errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def id(self):
        return self.pb.id

    @id.setter
    def id(self, value):
        if value is not None:
            self.pb.id = value

    @property
    def frequency(self):
        return self.pb.frequency if self.pb.HasField('frequency') else None

    @frequency.setter
    def frequency(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.frequency = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.frequency',
                                value=value,
                                metric_name='NetworkElement.Radio.frequency')
        else:
            self.pb.ClearField('frequency')

    @property
    def chan_width(self):
        return self.pb.chan_width if self.pb.HasField('chan_width') else None

    @chan_width.setter
    def chan_width(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.chan_width = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.chan_width',
                                value=value,
                                metric_name='NetworkElement.Radio.chan_width')
        else:
            self.pb.ClearField('chan_width')

    @property
    def tx_power(self):
        return self.pb.tx_power if self.pb.HasField('tx_power') else None

    @tx_power.setter
    def tx_power(self, value):
        if value is not None:
            if int.bit_length(value) < 32:
                self.pb.tx_power = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.tx_power',
                                value=value,
                                metric_name='NetworkElement.Radio.tx_power')
        else:
            self.pb.ClearField('tx_power')

    @property
    def dl_percent(self):
        return self.pb.dl_percent if self.pb.HasField('dl_percent') else None

    @dl_percent.setter
    def dl_percent(self, value):
        if value is not None:
            if value >= 0 and value <= 100:
                self.pb.dl_percent = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.dl_percent',
                                value=value,
                                metric_name='NetworkElement.Radio.dl_percent')
        else:
            self.pb.ClearField('dl_percent')

    @property
    def framing(self):
        return self.pb.framing

    @framing.setter
    def framing(self, value):
        if value is not None:
            self.pb.framing = value

    @property
    def rx_airtime(self):
        return self.pb.rx_airtime if self.pb.HasField('rx_airtime') else None

    @rx_airtime.setter
    def rx_airtime(self, value):
        if value is not None:
            if value >= 0:
                self.pb.rx_airtime = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.rx_airtime',
                                value=value,
                                metric_name='NetworkElement.Radio.rx_airtime')
        else:
            self.pb.ClearField('rx_airtime')

    @property
    def tx_airtime(self):
        return self.pb.tx_airtime if self.pb.HasField('tx_airtime') else None

    @tx_airtime.setter
    def tx_airtime(self, value):
        if value is not None:
            if value >= 0:
                self.pb.tx_airtime = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.tx_airtime',
                                value=value,
                                metric_name='NetworkElement.Radio.tx_airtime')
        else:
            self.pb.ClearField('tx_airtime')

    @property
    def gps_sync(self):
        return self.pb.gps_sync if self.pb.HasField('gps_sync') else None

    @gps_sync.setter
    def gps_sync(self, value):
        if value is not None:
            self.pb.gps_sync = value
        else:
            self.pb.ClearField('gps_sync')

    @property
    def chain_balance(self):
        return self.pb.chain_balance if self.pb.HasField('chain_balance') else None

    @chain_balance.setter
    def chain_balance(self, value):
        if value is not None:
            self.pb.chain_balance = value
        else:
            self.pb.ClearField('chain_balance')

    @property
    def rx_link_rate(self):
        return self.pb.rx_link_rate if self.pb.HasField('rx_link_rate') else None

    @rx_link_rate.setter
    def rx_link_rate(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.rx_link_rate = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.rx_link_rate',
                                value=value,
                                metric_name='NetworkElement.Radio.rx_link_rate')
        else:
            self.pb.ClearField('rx_link_rate')

    @property
    def tx_link_rate(self):
        return self.pb.tx_link_rate if self.pb.HasField('tx_link_rate') else None

    @tx_link_rate.setter
    def tx_link_rate(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.tx_link_rate = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.tx_link_rate',
                                value=value,
                                metric_name='NetworkElement.Radio.tx_link_rate')
        else:
            self.pb.ClearField('tx_link_rate')

    @property
    def frame_duration(self):
        return self.pb.frame_duration if self.pb.HasField('frame_duration') else None

    @frame_duration.setter
    def frame_duration(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.frame_duration = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.frame_duration',
                                value=value,
                                metric_name='NetworkElement.Radio.frame_duration')
        else:
            self.pb.ClearField('frame_duration')

    @property
    def rx_multiplexing_gain(self):
        return self.pb.rx_multiplexing_gain if self.pb.HasField(
            'rx_multiplexing_gain') else None

    @rx_multiplexing_gain.setter
    def rx_multiplexing_gain(self, value):
        if value is not None:
            self.pb.rx_multiplexing_gain = value
        else:
            self.pb.ClearField('rx_multiplexing_gain')

    @property
    def tx_multiplexing_gain(self):
        return self.pb.tx_multiplexing_gain if self.pb.HasField(
            'tx_multiplexing_gain') else None

    @tx_multiplexing_gain.setter
    def tx_multiplexing_gain(self, value):
        if value is not None:
            self.pb.tx_multiplexing_gain = value
        else:
            self.pb.ClearField('tx_multiplexing_gain')

    @property
    def rx_average_group_size(self):
        return self.pb.rx_average_group_size if self.pb.HasField(
            'rx_average_group_size') else None

    @rx_average_group_size.setter
    def rx_average_group_size(self, value):
        if value is not None:
            self.pb.rx_average_group_size = value
        else:
            self.pb.ClearField('rx_average_group_size')

    @property
    def tx_average_group_size(self):
        return self.pb.tx_average_group_size if self.pb.HasField(
            'tx_average_group_size') else None

    @tx_average_group_size.setter
    def tx_average_group_size(self, value):
        if value is not None:
            self.pb.tx_average_group_size = value
        else:
            self.pb.ClearField('tx_average_group_size')

    @property
    def rx_frutl(self):
        return self.pb.rx_frutl if self.pb.HasField('rx_frutl') else None

    @rx_frutl.setter
    def rx_frutl(self, value):
        if value is not None:
            if value >= 0:
                self.pb.rx_frutl = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.rx_frutl',
                                value=value,
                                metric_name='NetworkElement.Radio.rx_frutl')
        else:
            self.pb.ClearField('rx_frutl')

    @property
    def tx_frutl(self):
        return self.pb.tx_frutl if self.pb.HasField('tx_frutl') else None

    @tx_frutl.setter
    def tx_frutl(self, value):
        if value is not None:
            if value >= 0:
                self.pb.tx_frutl = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.tx_frutl',
                                value=value,
                                metric_name='NetworkElement.Radio.tx_frutl')
        else:
            self.pb.ClearField('tx_frutl')

    @property
    def wifi_mode(self):
        return self.pb.wifi_mode if self.pb.HasField('wifi_mode') else None

    @wifi_mode.setter
    def wifi_mode(self, value):
        if value is not None:
            if value > 0:
                self.pb.wifi_mode = value
            else:  # not a valid enum value
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                "Invalid value for Networkelement.Radio.wifi_mode",
                                value=value,
                                metric_name='NetworkElement.Radio.wifi_mode')
                self.pb.ClearField('wifi_mode')
        else:
            self.pb.ClearField('wifi_mode')

    @property
    def max_cpe_radios(self):
        return self.pb.max_cpe_radios if self.pb.HasField('max_cpe_radios') else None

    @max_cpe_radios.setter
    def max_cpe_radios(self, value):
        if value is None:
            self.pb.ClearField('max_cpe_radios')
        elif value >= 0 and int.bit_length(value) < 32:
            self.pb.max_cpe_radios = value
        elif self.errctx:
            self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                            'Invalid value for NetworkElement.Radio.max_cpe_radios',
                            value=value,
                            metric_name='NetworkElement.Radio.max_cpe_radios')

    @property
    def max_distance(self):
        return self.pb.max_distance if self.pb.HasField('max_distance') else None

    @max_distance.setter
    def max_distance(self, value):
        if value is None:
            self.pb.ClearField('max_distance')
        elif value >= 0 and int.bit_length(value) < 32:
            self.pb.max_distance = value
        elif self.errctx:
            self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                            'Invalid value for NetworkElement.Radio.max_distance',
                            value=value,
                            metric_name='NetworkElement.Radio.max_distance')

    @property
    def rx_path_loss(self):
        return self.pb.rx_path_loss if self.pb.HasField('rx_path_loss') else None

    @rx_path_loss.setter
    def rx_path_loss(self, value):
        if value is not None:
            if value >= -150 and value <= 150:
                self.pb.rx_path_loss = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Radio.rx_path_loss',
                                value=value,
                                metric_name='NetworkElement.Radio.rx_path_loss')
        else:
            self.pb.ClearField('rx_path_loss')

    @property
    def rx_max_interference_noise_ratio(self):
        return self.pb.rx_max_interference_noise_ratio if self.pb.HasField(
            'rx_max_interference_noise_ratio') else None

    @rx_max_interference_noise_ratio.setter
    def rx_max_interference_noise_ratio(self, value):
        if value is not None:
            if value >= -150 and value <= 150:
                self.pb.rx_max_interference_noise_ratio = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.Radio.rx_max_interference_noise_ratio',
                    value=value,
                    metric_name='NetworkElement.Radio.rx_max_interference_noise_ratio')
        else:
            self.pb.ClearField('rx_max_interference_noise_ratio')

    @property
    def rx_max_sensitivity_loss(self):
        return self.pb.rx_max_sensitivity_loss if self.pb.HasField(
            'rx_max_sensitivity_loss') else None

    @rx_max_sensitivity_loss.setter
    def rx_max_sensitivity_loss(self, value):
        if value is not None:
            if value >= -150 and value <= 150:
                self.pb.rx_max_sensitivity_loss = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.Radio.rx_max_sensitivity_loss',
                    value=value,
                    metric_name='NetworkElement.Radio.rx_max_sensitivity_loss')
        else:
            self.pb.ClearField('rx_max_sensitivity_loss')

    @property
    def azimuth(self):
        return self.pb.azimuth if self.pb.HasField('azimuth') else None

    @azimuth.setter
    def azimuth(self, value):
        if value is None:
            self.pb.ClearField('azimuth')
        elif value >= 0 and value <= 360:
            self.pb.azimuth = value
        elif self.errctx:
            self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                            'Invalid value for NetworkElement.Radio.azimuth',
                            value=value,
                            metric_name='NetworkElement.Radio.azimuth')

    @property
    def admin_up(self):
        return self.pb.admin_up if self.pb.HasField('admin_up') else None

    @admin_up.setter
    def admin_up(self, value):
        if value is not None:
            self.pb.admin_up = value
        else:
            self.pb.ClearField('admin_up')

    @property
    def oper_up(self):
        return self.pb.oper_up if self.pb.HasField('oper_up') else None

    @oper_up.setter
    def oper_up(self, value):
        if value is not None:
            self.pb.oper_up = value
        else:
            self.pb.ClearField('oper_up')


@pb_wrapper(network_poller_pb2.NetworkElement.Sector)
class Sector(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def ssid(self):
        return self.pb.ssid

    @ssid.setter
    def ssid(self, value):
        if value is not None:
            self.pb.ssid = value

    @property
    def mode(self):
        return self.pb.mode

    @mode.setter
    def mode(self, value):
        if value is not None:
            self.pb.mode = value


@pb_wrapper(network_poller_pb2.Error)
class Error(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)


@pb_wrapper(network_poller_pb2.NetworkElement.PonLink)
class PonLink(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def poller_hash(self):
        return self.pb.poller_hash

    @poller_hash.setter
    def poller_hash(self, value):
        if value is not None:
            self.pb.poller_hash = value

    @property
    def rx_power(self):
        return self.pb.rx_power if self.pb.HasField('rx_power') else None

    @rx_power.setter
    def rx_power(self, value):
        if value is not None:
            if value > -40 and value < 40:
                self.pb.rx_power = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.PonLink.rx_power',
                                value=value,
                                metric_name='NetworkElement.PonLink.rx_power')
        else:
            self.pb.ClearField('rx_power')

    @property
    def tx_power(self):
        return self.pb.tx_power if self.pb.HasField('tx_power') else None

    @tx_power.setter
    def tx_power(self, value):
        if value is not None:
            if value >= -50 and value <= 50:
                self.pb.tx_power = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.PonLink.tx_power',
                                value=value,
                                metric_name='NetworkElement.PonLink.tx_power')
        else:
            self.pb.ClearField('tx_power')

    @property
    def fiber_length(self):
        return self.pb.fiber_length if self.pb.HasField('fiber_length') else None

    @fiber_length.setter
    def fiber_length(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.fiber_length = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.PonLink.fiber_length',
                                value=value,
                                metric_name='NetworkElement.PonLink.fiber_length')
        else:
            self.pb.ClearField('fiber_length')

    @property
    def connected_time(self):
        return self.pb.connected_time if self.pb.HasField('connected_time') else None

    @connected_time.setter
    def connected_time(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.connected_time = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.PonLink.connected_time',
                    value=value,
                    metric_name='NetworkElement.PonLink.connected_time')
        else:
            self.pb.ClearField('connected_time')

    @property
    def rx_link_rate(self):
        return self.pb.rx_link_rate if self.pb.HasField('rx_link_rate') else None

    @rx_link_rate.setter
    def rx_link_rate(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.rx_link_rate = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.PonLink.rx_link_rate',
                                value=value,
                                metric_name='NetworkElement.PonLink.rx_link_rate')
        else:
            self.pb.ClearField('rx_link_rate')

    @property
    def tx_link_rate(self):
        return self.pb.tx_link_rate if self.pb.HasField('tx_link_rate') else None

    @tx_link_rate.setter
    def tx_link_rate(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.tx_link_rate = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.PonLink.tx_link_rate',
                                value=value,
                                metric_name='NetworkElement.PonLink.tx_link_rate')
        else:
            self.pb.ClearField('tx_link_rate')


@pb_wrapper(network_poller_pb2.NetworkElement.Pon)
class Pon(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        self.links = Repeated(PonLink, self.pb.links, self.errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def id(self):
        return self.pb.id

    @id.setter
    def id(self, value):
        if value is not None:
            self.pb.id = value

    @property
    def poller_hash(self):
        return self.pb.poller_hash

    @poller_hash.setter
    def poller_hash(self, value):
        if value is not None:
            self.pb.poller_hash = value

    @property
    def pon_technology(self):
        return self.pb.pon_technology if self.pb.HasField('pon_technology') else None

    @pon_technology.setter
    def pon_technology(self, value):
        if value is not None:
            if value > 0:
                self.pb.pon_technology = value
            else:  # not a valid enum value
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                "Invalid value for Networkelement.Pon.pon_technology",
                                value=value,
                                metric_name='NetworkElement.Pon.pon_technology')
                self.pb.ClearField('pon_technology')
        else:
            self.pb.ClearField('pon_technology')

    @property
    def split_ratio(self):
        return self.pb.split_ratio if self.pb.HasField('split_ratio') else None

    @split_ratio.setter
    def split_ratio(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.split_ratio = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Pon.split_ratio',
                                value=value,
                                metric_name='NetworkElement.Pon.split_ratio')
        else:
            self.pb.ClearField('split_ratio')

    @property
    def max_distance(self):
        return self.pb.max_distance if self.pb.HasField('max_distance') else None

    @max_distance.setter
    def max_distance(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.max_distance = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Pon.max_distance',
                                value=value,
                                metric_name='NetworkElement.Pon.max_distance')
        else:
            self.pb.ClearField('max_distance')

    @property
    def minimum_loss(self):
        return self.pb.minimum_loss if self.pb.HasField('minimum_loss') else None

    @minimum_loss.setter
    def minimum_loss(self, value):
        if value is not None:
            self.pb.minimum_loss = value
        else:
            self.pb.ClearField('minimum_loss')

    @property
    def maximum_loss(self):
        return self.pb.maximum_loss if self.pb.HasField('maximum_loss') else None

    @maximum_loss.setter
    def maximum_loss(self, value):
        if value is not None:
            self.pb.maximum_loss = value
        else:
            self.pb.ClearField('maximum_loss')

    @property
    def rx_link_rate(self):
        return self.pb.rx_link_rate if self.pb.HasField('rx_link_rate') else None

    @rx_link_rate.setter
    def rx_link_rate(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.rx_link_rate = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Pon.rx_link_rate',
                                value=value,
                                metric_name='NetworkElement.Pon.rx_link_rate')
        else:
            self.pb.ClearField('rx_link_rate')

    @property
    def tx_link_rate(self):
        return self.pb.tx_link_rate if self.pb.HasField('tx_link_rate') else None

    @tx_link_rate.setter
    def tx_link_rate(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.tx_link_rate = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Pon.tx_link_rate',
                                value=value,
                                metric_name='NetworkElement.Pon.tx_link_rate')
        else:
            self.pb.ClearField('tx_link_rate')

    @property
    def name(self):
        return self.pb.name if self.pb.HasField('name') else None

    @name.setter
    def name(self, value):
        if value is not None:
            self.pb.name = value
        else:
            self.pb.ClearField('name')

    @property
    def description(self):
        return self.pb.description if self.pb.HasField('description') else None

    @description.setter
    def description(self, value):
        if value is not None:
            self.pb.description = value
        else:
            self.pb.ClearField('description')


@pb_wrapper(network_poller_pb2.NetworkElement.Interface)
class Interface(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        self.ip_addresses = Repeated(None, self.pb.ip_addresses, self.errctx)
        self.bridged_mac_addresses = Repeated(None, self.pb.bridged_mac_addresses,
                                              self.errctx)
        self.links = Repeated(Link, self.pb.links, self.errctx)
        self.radios = Repeated(Radio, self.pb.radios, self.errctx)
        self.sectors = Repeated(Sector, self.pb.sectors, self.errctx)
        self.pons = Repeated(Pon, self.pb.pons, self.errctx)
        self.rf_aggregation_groups = Repeated(RfAggregationGroup,
                                              self.pb.rf_aggregation_groups,
                                              self.errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def id(self):
        return self.pb.id

    @id.setter
    def id(self, value):
        if value is not None:
            self.pb.id = value

    @property
    def name(self):
        return self.pb.name if self.pb.HasField('name') else None

    @name.setter
    def name(self, value):
        if value is not None:
            self.pb.name = value
        else:
            self.pb.ClearField('name')

    @property
    def type(self):
        return self.pb.type if self.pb.HasField('type') else None

    @type.setter
    def type(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.type = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Interface.type',
                                value=value,
                                metric_name='NetworkElement.Interface.type')
        else:
            self.pb.ClearField('type')

    @property
    def mtu(self):
        return self.pb.mtu if self.pb.HasField('mtu') else None

    @mtu.setter
    def mtu(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.mtu = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Interface.mtu',
                                value=value,
                                metric_name='NetworkElement.Interface.mtu')
        else:
            self.pb.ClearField('mtu')

    @property
    def speed(self):
        return self.pb.speed if self.pb.HasField('speed') else None

    @speed.setter
    def speed(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.speed = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Interface.speed',
                                value=value,
                                metric_name='NetworkElement.Interface.speed')
        else:
            self.pb.ClearField('speed')

    @property
    def mac_address(self):
        return self.pb.mac_address if self.pb.HasField('mac_address') else None

    @mac_address.setter
    def mac_address(self, value):
        if value is not None:
            self.pb.mac_address = value
        else:
            self.pb.ClearField('mac_address')

    @property
    def admin_up(self):
        return self.pb.admin_up if self.pb.HasField('admin_up') else None

    @admin_up.setter
    def admin_up(self, value):
        if value is not None:
            self.pb.admin_up = value
        else:
            self.pb.ClearField('admin_up')

    @property
    def oper_up(self):
        return self.pb.oper_up if self.pb.HasField('oper_up') else None

    @oper_up.setter
    def oper_up(self, value):
        if value is not None:
            self.pb.oper_up = value
        else:
            self.pb.ClearField('oper_up')

    @property
    def in_octets(self):
        return self.pb.in_octets if self.pb.HasField('in_octets') else None

    @in_octets.setter
    def in_octets(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_octets = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Interface.in_octets',
                                value=value,
                                metric_name='NetworkElement.Interface.in_octets')
        else:
            self.pb.ClearField('in_octets')

    @property
    def in_ucast_pkts(self):
        return self.pb.in_ucast_pkts if self.pb.HasField('in_ucast_pkts') else None

    @in_ucast_pkts.setter
    def in_ucast_pkts(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_ucast_pkts = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.Interface.in_ucast_pkts',
                    value=value,
                    metric_name='NetworkElement.Interface.in_ucast_pkts')
        else:
            self.pb.ClearField('in_ucast_pkts')

    @property
    def in_nucast_pkts(self):
        return self.pb.in_nucast_pkts if self.pb.HasField('in_nucast_pkts') else None

    @in_nucast_pkts.setter
    def in_nucast_pkts(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_nucast_pkts = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.Interface.in_nucast_pkts',
                    value=value,
                    metric_name='NetworkElement.Interface.in_nucast_pkts')
        else:
            self.pb.ClearField('in_nucast_pkts')

    @property
    def in_discards(self):
        return self.pb.in_discards if self.pb.HasField('in_discards') else None

    @in_discards.setter
    def in_discards(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_discards = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.Interface.in_discards',
                    value=value,
                    metric_name='NetworkElement.Interface.in_discards')
        else:
            self.pb.ClearField('in_discards')

    @property
    def in_errors(self):
        return self.pb.in_errors if self.pb.HasField('in_errors') else None

    @in_errors.setter
    def in_errors(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_errors = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Interface.in_errors',
                                value=value,
                                metric_name='NetworkElement.Interface.in_errors')
        else:
            self.pb.ClearField('in_errors')

    @property
    def in_unknown_protos(self):
        return self.pb.in_unknown_protos if self.pb.HasField(
            'in_unknown_protos') else None

    @in_unknown_protos.setter
    def in_unknown_protos(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.in_unknown_protos = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.Interface.in_unknown_protos',
                    value=value,
                    metric_name='NetworkElement.Interface.in_unknown_protos')
        else:
            self.pb.ClearField('in_unknown_protos')

    @property
    def out_octets(self):
        return self.pb.out_octets if self.pb.HasField('out_octets') else None

    @out_octets.setter
    def out_octets(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_octets = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Interface.out_octets',
                                value=value,
                                metric_name='NetworkElement.Interface.out_octets')
        else:
            self.pb.ClearField('out_octets')

    @property
    def out_ucast_pkts(self):
        return self.pb.out_ucast_pkts if self.pb.HasField('out_ucast_pkts') else None

    @out_ucast_pkts.setter
    def out_ucast_pkts(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_ucast_pkts = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.Interface.out_ucast_pkts',
                    value=value,
                    metric_name='NetworkElement.Interface.out_ucast_pkts')
        else:
            self.pb.ClearField('out_ucast_pkts')

    @property
    def out_nucast_pkts(self):
        return self.pb.out_nucast_pkts if self.pb.HasField('out_nucast_pkts') else None

    @out_nucast_pkts.setter
    def out_nucast_pkts(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_nucast_pkts = value
            elif self.errctx:
                self.errctx.add(
                    network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                    'Invalid value for NetworkElement.Interface.out_nucast_pkts',
                    value=value,
                    metric_name='NetworkElement.Interface.out_nucast_pkts')
        else:
            self.pb.ClearField('out_nucast_pkts')

    @property
    def out_discards(self):
        return self.pb.out_discards if self.pb.HasField('out_discards') else None

    @out_discards.setter
    def out_discards(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_discards = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Interface.discards',
                                value=value,
                                metric_name='NetworkElement.Interface.discards')
        else:
            self.pb.ClearField('out_discards')

    @property
    def out_errors(self):
        return self.pb.out_errors if self.pb.HasField('out_errors') else None

    @out_errors.setter
    def out_errors(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.out_errors = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Interface.out_errors',
                                value=value,
                                metric_name='NetworkElement.Interface.out_errors')
        else:
            self.pb.ClearField('out_errors')

    @property
    def poller_hash(self):
        return self.pb.poller_hash if self.pb.HasField('poller_hash') else None

    @poller_hash.setter
    def poller_hash(self, value):
        if value is not None:
            self.pb.poller_hash = value
        else:
            self.pb.ClearField('poller_hash')

    @property
    def manufacturer(self):
        return self.pb.manufacturer if self.pb.HasField('manufacturer') else None

    @manufacturer.setter
    def manufacturer(self, value):
        if value is not None:
            self.pb.manufacturer = value
        else:
            self.pb.ClearField('manufacturer')

    @property
    def model(self):
        return self.pb.model if self.pb.HasField('model') else None

    @model.setter
    def model(self, value):
        if value is not None:
            self.pb.model = value
        else:
            self.pb.ClearField('model')

    def add_bridged_mac(self, mac):
        if mac.startswith('33:33:') or mac == '00:00:00:00:00:00':
            return  # ignore IPv6 multicast MACs (PPA-1321) and zero MACs
        self.bridged_mac_addresses.append(mac)


@pb_wrapper(network_poller_pb2.IpMac)
class IpMac(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def ip_address(self):
        return self.pb.ip_address

    @ip_address.setter
    def ip_address(self, value):
        if value is not None:
            self.pb.ip_address = value

    @property
    def mac_address(self):
        return self.pb.mac_address

    @mac_address.setter
    def mac_address(self, value):
        if value is not None:
            self.pb.mac_address = value

    @property
    def port(self):
        return self.pb.port if self.pb.HasField('port') else None

    @port.setter
    def port(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.port = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElementUpdate.IpMac.port',
                                value=value,
                                metric_name='NetworkElementUpdate.IpMac.port')
        else:
            self.pb.ClearField('port')


@pb_wrapper(network_poller_pb2.Route)
class Route(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def net(self):
        return self.pb.net

    @net.setter
    def net(self, value):
        if value is not None:
            self.pb.net = value

    @property
    def if_index(self):
        return self.pb.if_index if self.pb.HasField('if_index') else None

    @if_index.setter
    def if_index(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.if_index = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElementUpdate.Route.if_index',
                                value=value,
                                metric_name='NetworkElementUpdate.Route.if_index')
        else:
            self.pb.ClearField('if_index')

    @property
    def next_hop(self):
        return self.pb.next_hop if self.pb.HasField('next_hop') else None

    @next_hop.setter
    def next_hop(self, value):
        if value is not None:
            self.pb.next_hop = value
        else:
            self.pb.ClearField('next_hop')

    @property
    def mac_address(self):
        return self.pb.mac_address if self.pb.HasField('mac_address') else None

    @mac_address.setter
    def mac_address(self, value):
        if value is not None:
            self.pb.mac_address = value
        else:
            self.pb.ClearField('mac_address')


@pb_wrapper(network_poller_pb2.DhcpLease)
class DhcpLease(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def ip_address(self):
        return self.pb.ip_address

    @ip_address.setter
    def ip_address(self, value):
        if value is not None:
            self.pb.ip_address = value

    @property
    def mac_address(self):
        return self.pb.mac_address

    @mac_address.setter
    def mac_address(self, value):
        if value is not None:
            self.pb.mac_address = value

    @property
    def circuit_id(self):
        return self.pb.circuit_id if self.pb.HasField('circuit_id') else None

    @circuit_id.setter
    def circuit_id(self, value):
        if value is not None:
            self.pb.circuit_id = value
        else:
            self.pb.ClearField('circuit_id')

    @property
    def remote_id(self):
        return self.pb.remote_id if self.pb.HasField('remote_id') else None

    @remote_id.setter
    def remote_id(self, value):
        if value is not None:
            self.pb.remote_id = value
        else:
            self.pb.ClearField('remote_id')

    @property
    def hostname(self):
        return self.pb.hostname if self.pb.HasField('hostname') else None

    @hostname.setter
    def hostname(self, value):
        if value is not None:
            self.pb.hostname = value
        else:
            self.pb.ClearField('hostname')


@pb_wrapper(network_poller_pb2.PppoeClient)
class PppoeClient(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        self.network_prefixes = Repeated(None, self.pb.network_prefixes, self.errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def username(self):
        return self.pb.username

    @username.setter
    def username(self, value):
        if value is not None:
            self.pb.username = value

    @property
    def mac_address(self):
        return self.pb.mac_address if self.pb.HasField('mac_address') else None

    @mac_address.setter
    def mac_address(self, value):
        if value is not None:
            self.pb.mac_address = value
        else:
            self.pb.ClearField('mac_address')


@pb_wrapper(network_poller_pb2.NetworkElement.Module)
class Module(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        self.interfaces = Repeated(Interface, self.pb.interfaces, self.errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def id(self):
        return self.pb.id

    @id.setter
    def id(self, value):
        if value is not None:
            self.pb.id = value

    @property
    def poller_hash(self):
        return self.pb.poller_hash

    @poller_hash.setter
    def poller_hash(self, value):
        if value is not None:
            self.pb.poller_hash = value

    @property
    def name(self):
        return self.pb.name if self.pb.HasField('name') else None

    @name.setter
    def name(self, value):
        if value is not None:
            self.pb.name = value
        else:
            self.pb.ClearField('name')

    @property
    def manufacturer(self):
        return self.pb.manufacturer if self.pb.HasField('manufacturer') else None

    @manufacturer.setter
    def manufacturer(self, value):
        if value is not None:
            self.pb.manufacturer = value
        else:
            self.pb.ClearField('manufacturer')

    @property
    def model(self):
        return self.pb.model if self.pb.HasField('model') else None

    @model.setter
    def model(self, value):
        if value is not None:
            self.pb.model = value
        else:
            self.pb.ClearField('model')

    @property
    def serial_number(self):
        return self.pb.serial_number if self.pb.HasField('serial_number') else None

    @serial_number.setter
    def serial_number(self, value):
        if value is not None:
            self.pb.serial_number = value
        else:
            self.pb.ClearField('serial_number')

    @property
    def sw_version(self):
        return self.pb.sw_version if self.pb.HasField('sw_version') else None

    @sw_version.setter
    def sw_version(self, value):
        if value is not None:
            self.pb.sw_version = value
        else:
            self.pb.ClearField('sw_version')

    @property
    def uptime(self):
        return self.pb.uptime if self.pb.HasField('uptime') else None

    @uptime.setter
    def uptime(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.uptime = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Module.uptime',
                                value=value,
                                metric_name='NetworkElement.Module.uptime')
        else:
            self.pb.ClearField('uptime')

    @property
    def cpu(self):
        return self.pb.cpu if self.pb.HasField('cpu') else None

    @cpu.setter
    def cpu(self, value):
        if value is not None:
            if value >= 0 and value <= 100:
                self.pb.cpu = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Module.cpu',
                                value=value,
                                metric_name='NetworkElement.Module.cpu')
        else:
            self.pb.ClearField('cpu')

    @property
    def mem(self):
        return self.pb.mem if self.pb.HasField('mem') else None

    @mem.setter
    def mem(self, value):
        if value is not None:
            if value >= 0 and value <= 100:
                self.pb.mem = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.Module.mem',
                                value=value,
                                metric_name='NetworkElement.Module.mem')
        else:
            self.pb.ClearField('mem')

    @property
    def temperature(self):
        return self.pb.temperature if self.pb.HasField('temperature') else None

    @temperature.setter
    def temperature(self, value):
        if value is not None:
            self.pb.temperature = value
        else:
            self.pb.ClearField('temperature')


@pb_wrapper(network_poller_pb2.NetworkElement)
class NetworkElement(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        self.modules = Repeated(Module, self.pb.modules, self.errctx)
        self.interfaces = Repeated(Interface, self.pb.interfaces, self.errctx)
        self.errors = Repeated(Error, self.pb.errors, self.errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def management_ip(self):
        return self.pb.management_ip if self.pb.HasField('management_ip') else None

    @management_ip.setter
    def management_ip(self, value):
        if value is not None:
            self.pb.management_ip = value
        else:
            self.pb.ClearField('management_ip')

    @property
    def name(self):
        return self.pb.name if self.pb.HasField('name') else None

    @name.setter
    def name(self, value):
        if value is not None:
            self.pb.name = value
        else:
            self.pb.ClearField('name')

    @property
    def manufacturer(self):
        return self.pb.manufacturer if self.pb.HasField('manufacturer') else None

    @manufacturer.setter
    def manufacturer(self, value):
        if value is not None:
            self.pb.manufacturer = value
        else:
            self.pb.ClearField('manufacturer')

    @property
    def model(self):
        return self.pb.model if self.pb.HasField('model') else None

    @model.setter
    def model(self, value):
        if value is not None:
            self.pb.model = value
        else:
            self.pb.ClearField('model')

    @property
    def serial_number(self):
        return self.pb.serial_number if self.pb.HasField('serial_number') else None

    @serial_number.setter
    def serial_number(self, value):
        if value is not None:
            self.pb.serial_number = value
        else:
            self.pb.ClearField('serial_number')

    @property
    def system_mac_address(self):
        return self.pb.system_mac_address if self.pb.HasField(
            'system_mac_address') else None

    @system_mac_address.setter
    def system_mac_address(self, value):
        if value is not None:
            if int(value.replace(':', ''), 16) < 2**48:
                self.pb.system_mac_address = value
            else:
                raise ValueError(f'Invalid MAC address "{value}"')
        else:
            self.pb.ClearField('system_mac_address')

    @property
    def sw_version(self):
        return self.pb.sw_version if self.pb.HasField('sw_version') else None

    @sw_version.setter
    def sw_version(self, value):
        if value is not None:
            self.pb.sw_version = value
        else:
            self.pb.ClearField('sw_version')

    @property
    def uptime(self):
        return self.pb.uptime if self.pb.HasField('uptime') else None

    @uptime.setter
    def uptime(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 64:
                self.pb.uptime = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.uptime',
                                value=value,
                                metric_name='NetworkElement.uptime')
        else:
            self.pb.ClearField('uptime')

    @property
    def cpu(self):
        return self.pb.cpu if self.pb.HasField('cpu') else None

    @cpu.setter
    def cpu(self, value):
        if value is not None:
            if value >= 0 and value <= 100:
                self.pb.cpu = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.cpu',
                                value=value,
                                metric_name='NetworkElement.cpu')
        else:
            self.pb.ClearField('cpu')

    @property
    def mem(self):
        return self.pb.mem if self.pb.HasField('mem') else None

    @mem.setter
    def mem(self, value):
        if value is not None:
            if value >= 0 and value <= 100:
                self.pb.mem = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.mem',
                                value=value,
                                metric_name='NetworkElement.mem')
        else:
            self.pb.ClearField('mem')

    @property
    def temperature(self):
        return self.pb.temperature if self.pb.HasField('temperature') else None

    @temperature.setter
    def temperature(self, value):
        if value is not None:
            self.pb.temperature = value
        else:
            self.pb.ClearField('temperature')

    @property
    def poller_hash(self):
        return self.pb.poller_hash

    @poller_hash.setter
    def poller_hash(self, value):
        if value is not None:
            self.pb.poller_hash = value

    @property
    def gps_sats_tracked(self):
        return self.pb.gps_sats_tracked if self.pb.HasField(
            'gps_sats_tracked') else None

    @gps_sats_tracked.setter
    def gps_sats_tracked(self, value):
        if value is not None:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.gps_sats_tracked = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElement.gps_sats_tracked',
                                value=value,
                                metric_name='NetworkElement.gps_sats_tracked')
        else:
            self.pb.ClearField('gps_sats_tracked')

    @property
    def latitude(self):
        return self.pb.latitude if self.pb.HasField('latitude') else None

    @latitude.setter
    def latitude(self, value):
        if value is not None:
            self.pb.latitude = value
        else:
            self.pb.ClearField('latitude')

    @property
    def longitude(self):
        return self.pb.longitude if self.pb.HasField('longitude') else None

    @longitude.setter
    def longitude(self, value):
        if value is not None:
            self.pb.longitude = value
        else:
            self.pb.ClearField('longitude')

    @property
    def elevation(self):
        return self.pb.elevation if self.pb.HasField('elevation') else None

    @elevation.setter
    def elevation(self, value):
        if value is not None:
            self.pb.elevation = value
        else:
            self.pb.ClearField('elevation')

    @property
    def height(self):
        return self.pb.height if self.pb.HasField('height') else None

    @height.setter
    def height(self, value):
        if value is not None:
            self.pb.height = value
        else:
            self.pb.ClearField('height')

    @property
    def status(self):
        return self.pb.status

    @status.setter
    def status(self, value):
        if value is not None:
            self.pb.status = value

    @property
    def imsi(self):
        return self.pb.imsi if self.pb.HasField('imsi') else None

    @imsi.setter
    def imsi(self, value):
        if value is None:
            self.pb.ClearField('imsi')
        else:
            self.pb.imsi = value

    @property
    def admin_disabled(self):
        return self.pb.admin_disabled if self.pb.HasField('admin_disabled') else None

    @admin_disabled.setter
    def admin_disabled(self, value):
        if value is None:
            self.pb.ClearField('admin_disabled')
        else:
            self.pb.admin_disabled = value

    @property
    def extra(self):
        return self.pb.extra if self.pb.HasField('extra') else None

    @extra.setter
    def extra(self, value):
        if value is not None:
            self.pb.extra = value
        else:
            self.pb.ClearField('extra')

    @property
    def counter_source(self):
        return self.pb.counter_source if self.pb.HasField('counter_source') else None

    @counter_source.setter
    def counter_source(self, value):
        if value is not None:
            self.pb.counter_source = value
        else:
            self.pb.ClearField('counter_source')

    @property
    def offline_reason(self):
        return self.pb.offline_reason if self.pb.HasField('offline_reason') else None

    @offline_reason.setter
    def offline_reason(self, value):
        if value is not None:
            if value > 0:
                self.pb.offline_reason = value
            else:  # not a valid enum value
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                "Invalid value for Networkelement.offline_reason",
                                value=value,
                                metric_name='NetworkElement.offline_reason')
                self.pb.ClearField('offline_reason')
        else:
            self.pb.ClearField('offline_reason')

    @property
    def offline_reason_detail(self):
        return self.pb.offline_reason_detail if self.pb.HasField(
            'offline_reason_detail') else None

    @offline_reason_detail.setter
    def offline_reason_detail(self, value):
        if value is not None:
            self.pb.offline_reason_detail = value
        else:
            self.pb.ClearField('offline_reason_detail')

    @property
    def offline_time(self):
        return self.pb.offline_time.ToDatetime(
            timezone.utc) if self.pb.HasField('offline_time') else None

    @offline_time.setter
    def offline_time(self, value):
        if value is not None:
            val = self.pb.offline_time.FromDatetime(value)
        else:
            self.pb.ClearField('offline_time')


@pb_wrapper(network_poller_pb2.NetworkElementUpdate)
class NetworkElementUpdate(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.errors = Repeated(None, self.pb.errors, self.errctx)
        self.peers = Repeated(NetworkElement, self.pb.peers, self.errctx)
        self.ip_macs = Repeated(IpMac, self.pb.ip_macs, self.errctx)
        self.routes = Repeated(Route, self.pb.routes, self.errctx)
        self.dhcp_leases = Repeated(DhcpLease, self.pb.dhcp_leases, self.errctx)
        self.pppoe_clients = Repeated(PppoeClient, self.pb.pppoe_clients, self.errctx)

    @property
    def time(self):
        return self.pb.time.seconds

    @time.setter
    def time(self, value):  # assumes seconds
        if value is not None:
            self.pb.time.FromSeconds(int(value))

    @property
    def company_uuid(self):
        return self.pb.company_uuid

    @company_uuid.setter
    def company_uuid(self, value):
        if value is not None:
            self.pb.company_uuid = value

    @property
    def element_uuid(self):
        return self.pb.element_uuid

    @element_uuid.setter
    def element_uuid(self, value):
        if value is not None:
            self.pb.element_uuid = value

    @property
    def instance(self):
        return self.pb.instance

    @instance.setter
    def instance(self, value):
        if value is not None:
            self.pb.instance = value

    @property
    def poller_version(self):
        return self.pb.poller_version

    @poller_version.setter
    def poller_version(self, value):
        if value is not None:
            self.pb.poller_version = value

    @property
    def active(self):
        return self.pb.active

    @active.setter
    def active(self, value):
        if value is not None:
            self.pb.active = value

    @property
    def pingable(self):
        return self.pb.pingable

    @pingable.setter
    def pingable(self, value):
        if value is not None:
            self.pb.pingable = value

    @property
    def ping_time(self):
        return self.pb.ping_time if self.pb.HasField('ping_time') else None

    @ping_time.setter
    def ping_time(self, value):
        if value is None:
            self.pb.ClearField('ping_time')
        else:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.ping_time = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElementUpdate.ping_time',
                                value=value,
                                metric_name='NetworkElementUpdate.ping_time')

    @property
    def snmp_version(self):
        return self.pb.snmp_version

    @snmp_version.setter
    def snmp_version(self, value):
        if value is not None:
            self.pb.snmp_version = value

    @property
    def sysobjid(self):
        return self.pb.sysobjid if self.pb.HasField('sysobjid') else None

    @sysobjid.setter
    def sysobjid(self, value):
        if value is None:
            self.pb.ClearField('sysobjid')
        else:
            self.pb.sysobjid = value

    @property
    def data(self):
        return NetworkElement(self.pb.data)

    @data.setter
    def data(self, value):
        if value is not None:
            self.pb.data.CopyFrom(value.pb)

    @property
    def poll_time(self):
        return self.pb.poll_time if self.pb.HasField('poll_time') else None

    @poll_time.setter
    def poll_time(self, value):
        if value is None:
            self.pb.ClearField('poll_time')
        elif value >= 0 and int.bit_length(value) < 32:
            self.pb.poll_time = value
        elif self.errctx:
            self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
                            'Invalid value for NetworkElementUpdate.poll_time',
                            value=value,
                            metric_name='NetworkElementUpdate.poll_time')

    @property
    def is_topology_stale(self):
        return self.pb.is_topology_stale if self.pb.HasField(
            'is_topology_stale') else None

    @is_topology_stale.setter
    def is_topology_stale(self, value):
        if value is not None:
            self.pb.is_topology_stale = value
        else:
            self.pb.ClearField('is_topology_stale')


@pb_wrapper(network_poller_pb2.NetworkElementDiscovery)
class NetworkElementDiscovery(BaseWrapperClass):

    def __init__(self, pb=None, errctx=None, **kwargs):
        super().__init__(pb, errctx)
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.errors = Repeated(None, self.pb.errors, self.errctx)

    @property
    def time(self):
        return self.pb.time.seconds

    @time.setter
    def time(self, value):  # assumes seconds
        if value is not None:
            self.pb.time.FromSeconds(int(value))

    @property
    def company_uuid(self):
        return self.pb.company_uuid

    @company_uuid.setter
    def company_uuid(self, value):
        if value is not None:
            self.pb.company_uuid = value

    @property
    def element_discovery_uuid(self):
        return self.pb.element_discovery_uuid

    @element_discovery_uuid.setter
    def element_discovery_uuid(self, value):
        if value is not None:
            self.pb.element_discovery_uuid = value

    @property
    def instance(self):
        return self.pb.instance

    @instance.setter
    def instance(self, value):
        if value is not None:
            self.pb.instance = value

    @property
    def active(self):
        return self.pb.active

    @active.setter
    def active(self, value):
        if value is not None:
            self.pb.active = value

    @property
    def pingable(self):
        return self.pb.pingable

    @pingable.setter
    def pingable(self, value):
        if value is not None:
            self.pb.pingable = value

    @property
    def ping_time(self):
        return self.pb.ping_time if self.pb.HasField('ping_time') else None

    @ping_time.setter
    def ping_time(self, value):
        if value is None:
            self.pb.ClearField('ping_time')
        else:
            if value >= 0 and int.bit_length(value) < 32:
                self.pb.ping_time = value
            elif self.errctx:
                self.errctx.add(network_poller_pb2.Error.ERROR_BAD_DATA_RANGE,
                                'Invalid value for NetworkElementDiscovery.ping_time',
                                value=value,
                                metric_name='NetworkElementDiscovery.ping_time')

    @property
    def snmp_version(self):
        return self.pb.snmp_version

    @snmp_version.setter
    def snmp_version(self, value):
        if value is not None:
            self.pb.snmp_version = value

    @property
    def sysobjid(self):
        return self.pb.sysobjid if self.pb.HasField('sysobjid') else None

    @sysobjid.setter
    def sysobjid(self, value):
        if value is None:
            self.pb.ClearField('sysobjid')
        else:
            self.pb.sysobjid = value

    @property
    def poller_version(self):
        return self.pb.poller_version

    @poller_version.setter
    def poller_version(self, value):
        if value is None:
            self.pb.ClearField('poller_version')
        else:
            self.pb.poller_version = value

    @property
    def poller_hash(self):
        return self.pb.poller_hash

    @poller_hash.setter
    def poller_hash(self, value):
        if value is not None:
            self.pb.poller_hash = value

    @property
    def status(self):
        return self.pb.status

    @status.setter
    def status(self, value):
        if value is not None:
            self.pb.status = value



"""Function to sync a Preseem network metadata store to the new Preseem model."""
import asyncio
from collections import namedtuple
from ipaddress import ip_address, ip_network
import logging
from uuid import UUID

from grpc import RpcError

from preseem import NetworkMetadataReference
from preseem.source_model import AccessPoint, Olt, Router, Service
from preseem_protobuf.model.common_pb2 import ElementStatus
from preseem_protobuf.model.service_pb2 import Attachment, SetAttachmentReq

# STM-6039 Only allow the sync to run once at a time, to prevent duplicate
# entities from being created.
sync_lock = None


# This object is used to pass element items to the sync_preseem_model
# algorithm for a given source and intent role.
class ElementSyncItem:

    def __init__(self, id, name, site, host):
        self.id = id
        self.name = name
        self.site = site
        self.host = host
        self._site_uuid = None

    def __repr__(self):
        clsname = self.__class__.__name__
        attrs = ', '.join(f"{k}={v!r}" for k, v in self.__dict__.items() if k[0] != '_')
        return f"{clsname}({attrs})"


_account_field_map = {
    'type': lambda x: x.type,
    'status': lambda x: x.status,
    'name': lambda x: x.name,
    'first_name': lambda x: x.first_name,
    'last_name': lambda x: x.last_name,
}

_package_field_map = {
    'name': lambda x: x.name,
    'download_rate': lambda x: x.download_rate or None,
    'upload_rate': lambda x: x.upload_rate or None,
}

_elem_disc_field_map = {
    'name': lambda x: x.name,
    'management_ip': lambda x: x.host,
    'site_uuid': lambda x: x._site_uuid,
    'intent_role_uuid': lambda x: x.role_uuid.bytes,
}

_service_field_map = {
    'account_uuid': lambda x: x._account_uuid,
    'package_uuid': lambda x: x._package_uuid,
    'download_rate': lambda x: x._download_rate or None,
    'upload_rate': lambda x: x._upload_rate or None,
    'attachment': lambda x: x._attachment.pb if x._attachment else None,
    'subscriber_identifier': lambda x: x._subscriber_identifier,
    'source_url': lambda x: x._url,
}

_elem_field_map = {
    'name': lambda x: x.name,
    'management_ip': lambda x: x.host,
    'site_uuid': lambda x: x._site_uuid,
    'intent_role_uuid': lambda x: x.role_uuid.bytes,
    'poller_hash': lambda x: x.poller_hash,
}


def has_any_fields_set(message):
    for field_desc, value in message.ListFields():
        if value:
            return True
    return False


async def call_and_log_error(co):
    """Call a GRPC coroutine and log any errors from it."""
    try:
        return await co
    except asyncio.CancelledError:
        raise
    except RpcError as err:
        logging.warning('sync_preseem_model: error calling service: %s (%s)',
                        err.code(), err.details())
    except Exception as err:
        logging.warning('sync_preseem_model: %s error calling service: %s',
                        type(err).__name__,
                        err,
                        exc_info=1)
    return None


async def run_limited(coros, limit: int):
    """Run a list of coroutines with a maximum concurrency limit."""
    sem = asyncio.Semaphore(limit)

    async def guarded(coro):
        async with sem:
            return await coro

    # schedule everything, but only `limit` run at a time
    return await asyncio.gather(*(guarded(c) for c in coros), )


async def _sync_objects(model,
                        company_uuid,
                        source,
                        bss_objs,
                        model_objs,
                        createrpc,
                        updaterpc,
                        field_map,
                        deactivate,
                        noact,
                        cfg=None,
                        init_fields=None,
                        concurrency=10):
    """
    Function to sync python bss-sourced objects to model objects.
    Pass in the model and the company_uuid and source we are working on.
    - bss_objs is a dict {source_id -> obj) where obj is a python bss object.
    - model_objs is a list of objects from the model.
    - createrpc and updaterpc are the grpc update methods for the object.
    - field_map is a dict of {model field name -> mapping func} where the mapping func
      returns the value to sync from the bss object.
    """
    new_ids = set(bss_objs)
    new_objs = []
    modcos = []
    delcos = []
    for mobj in model_objs:
        if mobj.source != source:
            continue  # ignore objects owned by other sources
        bobj = bss_objs.get(mobj.source_id)
        if bobj:
            changes = []
            mods = {}
            new_ids.discard(mobj.source_id)
            if mobj.inactive:
                changes.append('reactivated')
                mods['inactive'] = False
            for f, m in field_map.items():
                mval = getattr(mobj, f)
                bval = m(bobj)
                if mval != bval:
                    if f == 'attachment':
                        # find what attachment fields changed for logging purposes
                        batt = Attachment()
                        batt.CopyFrom(bval)  # make a copy so I don't modify the input
                        bval = batt
                        achgs = []
                        attachment_cfg = cfg.get('attachment') if cfg else {}
                        if mval and bval:
                            for sf in bval.DESCRIPTOR.fields:
                                smval = getattr(mval, sf.name)
                                sbval = getattr(bval, sf.name)
                                if sf.name == 'mac_addresses':
                                    smval = set(x for x in smval)
                                    sbval = set(x for x in sbval)
                                if sf.name == 'network_prefixes':
                                    smval = set([ip_network(x) for x in smval])
                                    sbval = set([ip_network(x) for x in sbval])
                                field_cfg = attachment_cfg.get(sf.name)
                                if field_cfg and field_cfg.get('ignore'):
                                    # To leave the field as it was, we have to set it
                                    # to what is in model-server now,
                                    # because the attachment is set as one full object.
                                    setattr(bval, sf.name, smval)
                                    sbval = smval
                                if smval != sbval:
                                    achgs.append(f"{sf.name}[{smval}->{sbval}]")
                        elif bval and not mval:
                            # PPA-318 the model sets an empty attachment to no
                            # attachment, so handle the comparison properly here.
                            if has_any_fields_set(bval):
                                achgs = [f"null->{bval}"]
                        elif mval and not bval:
                            achgs = [f"{bval}->null"]
                        if not achgs:
                            continue  # no further inspection, not actually a diff
                        changes.append(f'{f}: {",".join(achgs)}')
                    else:
                        changes.append(f'{f}: {mval} -> {bval}')
                    mods[f] = bval
            if mods:
                objdesc = mobj
                if isinstance(bobj, Service):
                    # more concise error message for Service
                    objdesc = f"Service {UUID(bytes=mobj.uuid)}"
                logging.info('sync_preseem_model(%s): modify %s (%s)', source, objdesc,
                             ', '.join(changes))
                if not noact:
                    if isinstance(bobj, Service) and 'attachment' in mods:
                        # special case for setting attachment
                        attachment = mods.pop('attachment')
                        set_att_req = SetAttachmentReq()
                        if attachment:
                            set_att_req.attachment.CopyFrom(attachment)
                        mods['attachment_req'] = set_att_req
                    modcos.append(
                        call_and_log_error(
                            updaterpc(company_uuid=company_uuid, uuid=mobj.uuid,
                                      **mods)))
        elif deactivate:  # sync deletes unless explicitly told not to
            if not mobj.inactive:
                logging.info("sync_preseem_model(%s): deactivate %s %s", source,
                             type(mobj).__name__.lower(), UUID(bytes=mobj.uuid))
                if not noact:
                    delcos.append(
                        call_and_log_error(
                            updaterpc(company_uuid=company_uuid,
                                      uuid=mobj.uuid,
                                      inactive=True)))

    # call modify and delete ops concurrently
    if modcos:
        logging.info(f"Modify {len(modcos)} objects")
        await run_limited(modcos, concurrency)
        logging.info(f"Modified {len(modcos)} objects")
    if delcos:
        logging.info(f"Deactivate {len(delcos)} objects")
        await run_limited(delcos, concurrency)
        logging.info(f"Deactivated {len(delcos)} objects")

    # Need to do these concurrent too, easy enough, will need a new helper fn.
    addcos = []
    for obj_id in new_ids:
        bobj = bss_objs.get(obj_id)
        if bobj:
            logging.info('sync_preseem_model(%s): create %s', source, bobj)
            if not noact:
                addcos.append(
                    call_and_log_error(
                        createrpc(company_uuid=company_uuid,
                                  source=source,
                                  source_id=bobj.id,
                                  **{
                                      f: m(bobj)
                                      for f, m in field_map.items()
                                  },
                                  **{
                                      f: v
                                      for f, v in (init_fields or {}).items()
                                  })))

    if addcos:
        logging.info(f"Create {len(addcos)} objects")
        mobjs = await run_limited(addcos, concurrency)
        for mobj in mobjs:
            if mobj:
                uuid = UUID(bytes=mobj.uuid)
                logging.info('sync_preseem_model(%s): created %s "%s"', source,
                             type(mobj).__name__, uuid)
                new_objs.append(mobj)
        logging.info(f"Created {len(addcos)} objects")
    return new_objs


async def _sync_sites(model, company_uuid, source, objs, noact=False):
    """Make sure model has all sites needed by source objs, and set the site_uuid on each obj."""
    sites = await model.Site.List(company_uuid=company_uuid)
    sites = {x.uuid: x for x in sites}

    # Create a map of name -> site.  Handle multiple sites with the same
    # name by preferring active sites, then taking the first sorted uuid.
    site_names = {}
    for site in sorted(sites.values(), key=lambda x: x.uuid):
        cur_site = site_names.get(site.name)
        if not cur_site or (cur_site.inactive and not site.inactive):
            site_names[site.name] = site

    new_sites = set()
    for sobj in objs:
        if sobj.site:
            site = site_names.get(sobj.site)
            if not site:
                new_sites.add(sobj.site)
            elif site.inactive:
                logging.info('sync_sites(%s): reactivate site "%s"', source, sobj.site)
                if not noact:
                    await call_and_log_error(
                        model.Site.Update(company_uuid=company_uuid,
                                          uuid=site.uuid,
                                          inactive=False))
    for site_name in new_sites:
        logging.info('sync_sites(%s): create site "%s"', source, site_name)
        if not noact:
            site = site_names[site_name] = await call_and_log_error(
                model.Site.Create(company_uuid=company_uuid, name=site_name))
            if site:
                sites[site.uuid] = site_names[site_name] = site

    for sobj in objs:
        if sobj.site:
            site = site_names.get(sobj.site)
            sobj._site_uuid = site.uuid
        else:
            sobj._site_uuid = None


async def sync_preseem_model(preseem_model,
                             company_uuid,
                             source,
                             role,
                             items,
                             packages=None,
                             accounts=None,
                             services=None,
                             routers=None,
                             olts=None,
                             aps=None,
                             nm_model=None,
                             deactivate=None,
                             noact=False,
                             cfg=None,
                             verbose=0):
    """
    Update the Preseem model to keep its ElementDiscovery objects in sync with
    another system.  This method takes the full set of items for a specific
    company, source, and role.  It diffs this against the current model and
    applies any create, modify or delete (set inactive) operations.
    It returns a map of {id -> uuid} for the caller to use.
    """
    global sync_lock
    if sync_lock is None:
        sync_lock = asyncio.Lock()

    assert isinstance(company_uuid, bytes)
    assert isinstance(source, str)
    if role:
        # PPA-1512 pass role=None to disable the old AP sync method.
        assert isinstance(role, UUID)

    if source == '':
        raise ValueError('source cannot be the empty string')

    async with sync_lock:
        # Collect all discovery objects so we can sync the sites.
        ne_objs = [
            obj for d in (aps, routers, olts) if d is not None for obj in d.values()
        ]
        for item in items or []:
            if item.site is not None and (not isinstance(item.site, str)
                                          or item.site == ''):
                raise ValueError('invalid string value for site')
            ne_objs.append(item)

        model_accounts, model_discs, model_services, model_packages, _ = await asyncio.gather(
            preseem_model.Account.List(company_uuid=company_uuid),
            preseem_model.Element.ListElementDiscovery(company_uuid=company_uuid),
            preseem_model.Service.List(company_uuid=company_uuid,
                                       include_attachment=True),
            preseem_model.Package.List(company_uuid=company_uuid),
            _sync_sites(preseem_model, company_uuid, source, ne_objs, noact))

        role_discs = {}
        for disc in model_discs:
            discs = role_discs.get(disc.intent_role_uuid)
            if discs is None:
                discs = role_discs[disc.intent_role_uuid] = []
            discs.append(disc)

        if role:
            model_discs = {x.uuid: x for x in role_discs.get(role.bytes) or []}
            disc_ids = {(x.source, x.source_id): x for x in model_discs.values()}

        # Make sure all sites we reference exist and do some pre-validation.
        new_sites = set()
        item_ids = set()
        dup_ids = set()
        for item in items:
            if not isinstance(item.id, str) or item.id == '':
                raise ValueError('invalid string value for source_id')
            if not isinstance(item.name, str) or item.name == '':
                raise ValueError('invalid string value for name')
            if item.id in item_ids:
                dup_ids.add(item.id)
            else:
                item_ids.add(item.id)
        if dup_ids:
            logging.warning('sync_preseem_model(%s): duplicate IDs found: %s', source,
                            dup_ids)
            raise ValueError('duplicate IDs found')

        uuid_map = {}
        if role:  # old algorithm
            unref_discs = set(
                [x[0] for x in model_discs.items() if x[1].source == source])
            for item in items:
                host = str(ip_address(item.host))  # validate ip address
                key = (source, item.id)
                disc = disc_ids.get(key)
                if disc:
                    uuid_map[item.id] = UUID(bytes=disc.uuid)
                    unref_discs.remove(disc.uuid)
                    changes = []
                    mods = {}
                    if disc.inactive:
                        changes.append('reactivated')
                        mods['inactive'] = False
                        mods['element_uuid'] = None  # STM-9823
                    if disc.name != item.name:
                        changes.append(f'name: {disc.name} -> {item.name}')
                        mods['name'] = item.name
                    if disc.management_ip != host:
                        changes.append(f'management_ip: {disc.management_ip} -> {host}')
                        mods['management_ip'] = host
                    if disc.site_uuid != item._site_uuid:
                        changes.append(f'site: {disc.site_uuid} -> {item._site_uuid}')
                        mods['site_uuid'] = item._site_uuid
                    if mods:
                        logging.info('sync_preseem_model(%s): modify %s (%s)', source,
                                     item, ','.join(changes))
                        if not noact:
                            if nm_model and 'reactivated' in changes and disc.element_uuid:
                                # STM-9823 as part of clearing the element_uuid,
                                # we also need to delete the elem_discovery ref.
                                ref = NetworkMetadataReference(
                                    'elem_discovery', str(UUID(bytes=disc.uuid)), {})
                                logging.info(' delete reference: %s', ref)
                                await nm_model.del_ref(ref)
                            await call_and_log_error(
                                preseem_model.Element.UpdateElementDiscovery(
                                    company_uuid=company_uuid, uuid=disc.uuid, **mods))

                else:
                    logging.info('sync_preseem_model(%s): create %s', source, item)
                    if not noact:
                        disc = await call_and_log_error(
                            preseem_model.Element.CreateElementDiscovery(
                                company_uuid=company_uuid,
                                name=item.name,
                                management_ip=item.host,
                                site_uuid=item._site_uuid,
                                source=source,
                                source_id=item.id,
                                intent_role_uuid=role.bytes))
                        if disc:
                            uuid = UUID(bytes=disc.uuid)
                            logging.info(
                                'sync_preseem_model(%s): created ElementDiscovery "%s"',
                                source, uuid)
                            uuid_map[item.id] = uuid
            for disc_uuid in unref_discs:
                disc = model_discs.get(disc_uuid)
                if not disc.inactive:
                    logging.info("sync_preseem_model(%s): deactivate %s", source,
                                 UUID(bytes=disc_uuid))
                    if not noact:
                        await call_and_log_error(
                            preseem_model.Element.UpdateElementDiscovery(
                                company_uuid=company_uuid,
                                uuid=disc_uuid,
                                inactive=True))
        new_packages = []
        if packages is not None:
            assert isinstance(packages, dict)
            for pkg in packages.values():
                if not isinstance(pkg.id, str) or pkg.id == '':
                    raise ValueError('invalid string value for package source_id')
                if not isinstance(pkg.name, str) or pkg.name == '':
                    raise ValueError('invalid string value for package name')
                if pkg.download_rate and (not isinstance(pkg.download_rate, int)
                                          or pkg.download_rate < 0):
                    raise ValueError('invalid value for package down rate')
                if pkg.upload_rate and (not isinstance(pkg.upload_rate, int)
                                        or pkg.upload_rate < 0):
                    raise ValueError('invalid value for package up rate')
            new_packages = await _sync_objects(preseem_model,
                                               company_uuid,
                                               source,
                                               packages,
                                               model_packages,
                                               preseem_model.Package.Create,
                                               preseem_model.Package.Update,
                                               _package_field_map,
                                               deactivate=deactivate is not False,
                                               noact=noact)

        new_accounts = []
        if accounts is not None:
            assert isinstance(accounts, dict)
            for account in accounts.values():
                if not isinstance(account.id, str) or account.id == '':
                    raise ValueError('invalid string value for account source_id')
                if account.type is not None and (not isinstance(account.type, int)
                                                 or account.type <= 0):
                    raise ValueError('invalid value for account type')
                if account.status is None or not isinstance(account.status,
                                                            int) or account.status <= 0:
                    raise ValueError('invalid value for account status')
                if account.name is not None and (not isinstance(account.name, str)
                                                 or account.name == ''):
                    raise ValueError('invalid string value for account name')
                if account.first_name is not None and (not isinstance(
                        account.first_name, str) or account.first_name == ''):
                    raise ValueError('invalid string value for account first_name')
                if account.last_name is not None and (not isinstance(
                        account.last_name, str) or account.last_name == ''):
                    raise ValueError('invalid string value for account last_name')
            new_accounts = await _sync_objects(preseem_model,
                                               company_uuid,
                                               source,
                                               accounts,
                                               model_accounts,
                                               preseem_model.Account.Create,
                                               preseem_model.Account.Update,
                                               _account_field_map,
                                               deactivate=deactivate is not False,
                                               noact=noact)

        if services is not None:
            assert isinstance(services, dict)
            # Setup maps to lookup account and package UUIDs from the source IDs.
            # We have to handle cases where there are multiple matching objects.
            accounts_by_id = {}
            for account in model_accounts + new_accounts:
                al = accounts_by_id.get(account.source_id)
                if al is None:
                    accounts_by_id[account.source_id] = account
                elif isinstance(al, list):
                    al.append(account)
                else:
                    accounts_by_id[account.source_id] = [al, account]
            packages_by_id = {}
            for package in model_packages + new_packages:
                pl = packages_by_id.get(package.source_id)
                if pl is None:
                    packages_by_id[package.source_id] = package
                elif isinstance(pl, list):
                    pl.append(package)
                else:
                    packages_by_id[package.source_id] = [pl, package]

            # type/value checks are done in the source_model class
            for service in services.values():
                assert isinstance(service, Service)
                account = accounts_by_id.get(
                    service.account_id) if service.account_id else None
                if account:
                    if isinstance(account, list):  # multiple matching accounts
                        acc = next((x for x in account if x.source == source), None)
                        if acc:
                            account = acc  # use account with matching source
                        else:
                            account = None  # multiple non-source matches; raise error
                            service.err(
                                'account_id', service.account_id,
                                ValueError("Multiple accounts match the account ID"))
                    if account:
                        service._account_uuid = account.uuid
                package = packages_by_id.get(
                    service.package_id) if service.package_id else None
                if package:
                    if isinstance(package, list):  # multiple matching packages
                        pkg = next((x for x in package if x.source == source), None)
                        if pkg:
                            package = pkg  # use package with matching source
                        else:
                            package = None  # multiple non-source matches; raise error
                            service.err(
                                'package_id', service.package_id,
                                ValueError("Multiple packages match the package ID"))
                    if package:
                        service._package_uuid = package.uuid
            await _sync_objects(preseem_model,
                                company_uuid,
                                source,
                                services,
                                model_services,
                                preseem_model.Service.Create,
                                preseem_model.Service.Update,
                                _service_field_map,
                                deactivate=deactivate is not False,
                                cfg=cfg.get('service') if cfg else None,
                                noact=noact)

        for bss_objs, source_type in (aps, AccessPoint), (routers, Router), (olts, Olt):
            if bss_objs is not None:
                assert isinstance(bss_objs, dict)
                for sobj in bss_objs.values():
                    assert isinstance(sobj, source_type)
                    if not sobj.name:  # name is required for ElementDiscovery
                        raise ValueError(
                            f'invalid value for {source_type.typename} name')
                    if not sobj.host:  # host is required for ElementDiscovery
                        raise ValueError(
                            f'invalid value for {source_type.typename} host')

                discs = [
                    x for x in role_discs.get(source_type.role_uuid.bytes) or []
                    if x.source == source
                ]
                new_objs = await _sync_objects(
                    preseem_model,
                    company_uuid,
                    source,
                    bss_objs,
                    discs,
                    preseem_model.Element.CreateElementDiscovery,
                    preseem_model.Element.UpdateElementDiscovery,
                    _elem_disc_field_map,
                    deactivate=deactivate is not False,
                    noact=noact)
                for mobj in discs + new_objs:
                    sobj = bss_objs.get(mobj.source_id)
                    if sobj:
                        sobj._uuid = mobj.uuid

        return uuid_map


async def sync_preseem_elements(model, company_uuid, source, objs, noact=False):
    """
    Update Elements directly in the Preseem model, given source objects such as APs,
    OLTs, Routers.  This is to support cases when another system such as an NMS is the
    source of truth for an element's data, and we will not be polling the element
    directory.  In this case, the element may not even have (or we may not know) a 
    management IP address, and there is no point in having the poller discover it.
    """
    assert isinstance(company_uuid, bytes)
    assert isinstance(source, str)
    if source == '':
        raise ValueError('source cannot be the empty string')

    global sync_lock
    if sync_lock is None:
        sync_lock = asyncio.Lock()

    async with sync_lock:
        # It would be more efficient if we had a way to get elements by source,
        # but there is no such method, so we get all elements and filter what we need.
        all_elems, _ = await asyncio.gather(
            model.Element.List(company_uuid=company_uuid),
            _sync_sites(model, company_uuid, source, objs.values(), noact=noact))

        # Only keep elements that match a source and poller hash in objs.
        # This works because we don't currently disable elements anyway, and it
        # makes sure we're syncing the correct element in the case of a hardware swap.
        phs = {x.poller_hash: x for x in objs.values()}

        elems = []
        for elem in all_elems:
            phelem = phs.get(elem.poller_hash)
            if elem.source != source or not phelem:
                continue
            elems.append(elem)

            # Handle ID changes with the same poller hash.  This needs to be the same
            # element object.
            if phelem and phelem.id != elem.source_id:
                logging.info(
                    "sync_preseem_element: change element source_id %s: %s -> %s",
                    UUID(bytes=elem.uuid), elem.source_id, phelem.id)
                if not noact:
                    await call_and_log_error(
                        model.Element.Update(company_uuid=company_uuid,
                                             uuid=elem.uuid,
                                             source_id=phelem.id))
                elem.source_id = phelem.id

        # Sync the objects to the model.  New elements need their status set.
        init_fields = {'status': ElementStatus.ELEMENT_STATUS_UNKNOWN}
        new_objs = await _sync_objects(model,
                                       company_uuid,
                                       source,
                                       objs,
                                       elems,
                                       model.Element.Create,
                                       model.Element.Update,
                                       _elem_field_map,
                                       deactivate=False,
                                       init_fields=init_fields,
                                       noact=noact)
        for mobj in elems + new_objs:
            sobj = objs.get(mobj.source_id)
            if sobj:
                sobj._uuid = mobj.uuid
